"""pyvizio utility constants."""

APK_SOURCE_PATH = "src"
RESOURCE_PATH = "resources/res/raw"
APP_NAMES_FILE = "apps.json"
APP_PAYLOADS_FILE = "apps_availability.json"
APP_NAMES_URL = "http://hometest.buddytv.netdna-cdn.com/appservice/vizio_apps_prod.json"
APP_PAYLOADS_URL = (
    "http://hometest.buddytv.netdna-cdn.com/appservice/app_availability_prod.json"
)



"""Initialize pyvizio."""
import logging
from typing import Any, Dict, List, Optional, Union
from urllib.parse import urlsplit

from aiohttp import ClientSession
import requests
import xmltodict

from pyvizio.api._protocol import KEY_CODE, async_invoke_api, async_invoke_api_auth
from pyvizio.api.apps import (
    AppConfig,
    GetCurrentAppConfigCommand,
    GetCurrentAppNameCommand,
    LaunchAppConfigCommand,
    LaunchAppNameCommand,
)
from pyvizio.api.base import CommandBase
from pyvizio.api.input import (
    ChangeInputCommand,
    GetCurrentInputCommand,
    GetInputsListCommand,
    InputItem,
)
from pyvizio.api.item import (
    GetCurrentPowerStateCommand,
    GetDeviceInfoCommand,
    GetESNCommand,
    GetModelNameCommand,
    GetSerialNumberCommand,
    GetVersionCommand,
)
from pyvizio.api.pair import (
    BeginPairCommand,
    BeginPairResponse,
    CancelPairCommand,
    PairChallengeCommand,
    PairChallengeResponse,
)
from pyvizio.api.remote import EmulateRemoteCommand
from pyvizio.api.settings import (
    ChangeSettingCommand,
    GetAllSettingsCommand,
    GetAllSettingsOptionsCommand,
    GetAllSettingsOptionsXListCommand,
    GetAllSettingTypesCommand,
    GetSettingCommand,
    GetSettingOptionsCommand,
    GetSettingOptionsXListCommand,
)
from pyvizio.const import (
    APP_HOME,
    APPS,
    DEFAULT_DEVICE_CLASS,
    DEFAULT_PORTS,
    DEFAULT_TIMEOUT,
    DEVICE_CLASS_SPEAKER,
    DEVICE_CLASS_TV,
    MAX_VOLUME,
)
from pyvizio.discovery.ssdp import SSDPDevice, discover as discover_ssdp
from pyvizio.discovery.zeroconf import ZeroconfDevice, discover as discover_zc
from pyvizio.helpers import async_to_sync, open_port
from pyvizio.util import gen_apps_list_from_url

_LOGGER = logging.getLogger(__name__)


class VizioAsync:
    """Asynchronous class to interact with Vizio SmartCast devices."""

    def __init__(
        self,
        device_id: str,
        ip: str,
        name: str,
        auth_token: str = "",
        device_type: str = DEFAULT_DEVICE_CLASS,
        session: Optional[ClientSession] = None,
        timeout: int = DEFAULT_TIMEOUT,
    ) -> None:
        """Initialize asynchronous class to interact with Vizio SmartCast devices."""
        self.device_type = device_type.lower()
        if (
            self.device_type != DEVICE_CLASS_TV
            and self.device_type != DEVICE_CLASS_SPEAKER  # noqa: W503
        ):
            raise Exception(
                f"Invalid device type specified. Use either '{DEVICE_CLASS_TV}' or "
                f"'{DEVICE_CLASS_SPEAKER}'"
            )

        self._auth_token = auth_token
        self.ip = ip
        self.name = name
        self.device_id = device_id
        self._session = session
        self._timeout = timeout

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.__dict__})"

    def __eq__(self, other) -> bool:
        return self is other or self.__dict__ == other.__dict__

    async def __add_port(self) -> None:
        """Asynchronously add first open port from known ports list to `ip` property."""
        for port in DEFAULT_PORTS:
            if await open_port(self.ip, port):
                self.ip = f"{self.ip}:{port}"

    async def __invoke_api(
        self, cmd: CommandBase, log_api_exception: bool = True
    ) -> Any:
        """Asynchronously call SmartCast API without auth token."""
        if ":" not in self.ip:
            await self.__add_port()

        return await async_invoke_api(
            self.ip,
            cmd,
            _LOGGER,
            custom_timeout=self._timeout,
            log_api_exception=log_api_exception,
            session=self._session,
        )

    async def __invoke_api_auth(
        self, cmd: CommandBase, log_api_exception: bool = True
    ) -> Any:
        """Asynchronously call SmartCast API with auth token."""
        if ":" not in self.ip:
            await self.__add_port()

        return await async_invoke_api_auth(
            self.ip,
            cmd,
            _LOGGER,
            auth_token=self._auth_token,
            custom_timeout=self._timeout,
            log_api_exception=log_api_exception,
            session=self._session,
        )

    async def __invoke_api_may_need_auth(
        self, cmd: CommandBase, log_api_exception: bool = True
    ) -> Any:
        """Asynchronously call SmartCast API command with or without auth token depending on device type."""
        if not self._auth_token:
            if self.device_type == DEVICE_CLASS_SPEAKER:
                return await self.__invoke_api(cmd, log_api_exception=log_api_exception)
            else:
                raise Exception(
                    f"Empty auth token. To target a speaker and bypass auth "
                    f"requirements, pass '{DEVICE_CLASS_SPEAKER}' as device_type"
                )
        return await self.__invoke_api_auth(cmd, log_api_exception=log_api_exception)

    async def __remote(
        self, key_list: Union[str, List[str]], log_api_exception: bool = True
    ) -> bool:
        """Asynchronously call key press API with list of keys."""
        key_codes = []
        if not isinstance(key_list, list):
            key_list = [key_list]

        for key in key_list:
            if key not in KEY_CODE[self.device_type]:
                _LOGGER.error(
                    "Key Code of '%s' not found for device type of '%s'",
                    key,
                    self.device_type,
                )
                return False
            else:
                key_codes.append(KEY_CODE[self.device_type][key])

        result = await self.__invoke_api_may_need_auth(
            EmulateRemoteCommand(key_codes, self.device_type),
            log_api_exception=log_api_exception,
        )
        return result is not None

    async def __remote_multiple(
        self, key_code: str, num: int, log_api_exception: bool = True
    ) -> bool:
        """Asynchronously call key press API with list of same key repeated multiple times."""
        key_codes = [key_code for _ in range(num)]
        return await self.__remote(key_codes, log_api_exception=log_api_exception)

    @staticmethod
    def discovery_zeroconf(timeout: int = DEFAULT_TIMEOUT) -> List[ZeroconfDevice]:
        """Discover Vizio devices on network using zeroconf."""
        results = discover_zc("_viziocast._tcp.local.", timeout=timeout)
        _LOGGER.info(results)
        return results

    @staticmethod
    def discovery_ssdp(timeout: int = DEFAULT_TIMEOUT) -> List[SSDPDevice]:
        """Discover Vizio devices on network using SSDP."""
        results = []

        devices = discover_ssdp(
            "urn:dial-multiscreen-org:device:dial:1", timeout=timeout
        )

        for dev in devices:
            data = xmltodict.parse(
                requests.get(dev.location, verify=False, timeout=timeout).text
            )

            if "root" not in data or "device" not in data["root"]:
                continue

            root = data["root"]["device"]
            manufacturer = root["manufacturer"]

            if manufacturer is None or "VIZIO" != manufacturer:
                continue

            split_url = urlsplit(dev.location)
            device = SSDPDevice(
                split_url.hostname, root["friendlyName"], root["modelName"], root["UDN"]
            )

            results.append(device)
        _LOGGER.info(results)
        return results

    @staticmethod
    async def validate_ha_config(
        ip: str,
        auth_token: str,
        device_type: str,
        session: Optional[ClientSession] = None,
        timeout: int = DEFAULT_TIMEOUT,
    ) -> bool:
        """Asynchronously return whether or not HomeAssistant config will allow HomeAssistant to make successful calls to Vizio SmartCast API."""
        return await VizioAsync(
            "", ip, "", auth_token, device_type, session=session, timeout=timeout
        ).can_connect_with_auth_check()

    @staticmethod
    async def get_unique_id(
        ip: str,
        device_type: str,
        timeout: int = DEFAULT_TIMEOUT,
        session: Optional[ClientSession] = None,
    ) -> Optional[str]:
        """Asynchronously get unique identifier for Vizio device."""
        return await VizioAsync(
            "", ip, "", "", device_type, session=session, timeout=timeout
        ).get_serial_number(log_api_exception=False)

    async def can_connect_with_auth_check(self) -> bool:
        """Asynchronously return whether or not device API can be connected to with valid authorization."""
        return bool(
            await VizioAsync.get_all_audio_settings(self, log_api_exception=False)
        )

    async def can_connect_no_auth_check(self) -> bool:
        """Asynchronously return whether or not device API can be connected to regardless of authorization."""
        return bool(
            await self.__invoke_api(
                GetDeviceInfoCommand(self.device_type), log_api_exception=False
            )
        )

    async def get_esn(self, log_api_exception: bool = True) -> Optional[str]:
        """Asynchronously get device's ESN (electronic serial number?)."""
        item = await self.__invoke_api_may_need_auth(
            GetESNCommand(self.device_type), log_api_exception=log_api_exception
        )

        if item:
            return item.value

        return None

    async def get_serial_number(self, log_api_exception: bool = True) -> Optional[str]:
        """Asynchronously get device's serial number."""
        item = await self.__invoke_api(
            GetSerialNumberCommand(self.device_type),
            log_api_exception=log_api_exception,
        )

        if item:
            return item.value

        return None

    async def get_version(self, log_api_exception: bool = True) -> Optional[str]:
        """Asynchronously get SmartCast software version on device."""
        item = await self.__invoke_api(
            GetVersionCommand(self.device_type), log_api_exception=log_api_exception
        )

        if item:
            return item.value

        return None

    async def get_model_name(self, log_api_exception: bool = True) -> Optional[str]:
        """Asynchronously get device's model number."""
        return await self.__invoke_api(
            GetModelNameCommand(self.device_type), log_api_exception=log_api_exception
        )

    async def start_pair(
        self, log_api_exception: bool = True
    ) -> Optional[BeginPairResponse]:
        """Asynchronously begin pairing process to obtain challenge type and challenge token."""
        return await self.__invoke_api(
            BeginPairCommand(self.device_id, self.name, self.device_type),
            log_api_exception=log_api_exception,
        )

    async def stop_pair(self, log_api_exception: bool = True) -> Optional[bool]:
        """Asynchronously cancel pairing process."""
        return await self.__invoke_api(
            CancelPairCommand(self.device_id, self.name, self.device_type),
            log_api_exception=log_api_exception,
        )

    async def pair(
        self, ch_type: int, token: int, pin: str = "", log_api_exception: bool = True
    ) -> Optional[PairChallengeResponse]:
        """Asynchronously complete pairing process to obtain auth token."""
        if self.device_type == DEVICE_CLASS_SPEAKER:
            pin = "0000"
        return await self.__invoke_api(
            PairChallengeCommand(self.device_id, ch_type, token, pin, self.device_type),
            log_api_exception=log_api_exception,
        )

    async def get_inputs_list(
        self, log_api_exception: bool = True
    ) -> Optional[List[InputItem]]:
        """Asynchronously get list of available inputs."""
        return await self.__invoke_api_may_need_auth(
            GetInputsListCommand(self.device_type), log_api_exception=log_api_exception
        )

    async def get_current_input(self, log_api_exception: bool = True) -> Optional[str]:
        """Asynchronously get device's active input."""
        item = await self.__invoke_api_may_need_auth(
            GetCurrentInputCommand(self.device_type),
            log_api_exception=log_api_exception,
        )

        if item:
            return item.meta_name

        return None

    async def next_input(self, log_api_exception: bool = True) -> Optional[bool]:
        """Asynchronously switch active input to next input."""
        # HACK: Single call just invoking overlay menu with current input
        return await self.__remote_multiple(
            "INPUT_NEXT", 2, log_api_exception=log_api_exception
        )

    async def set_input(
        self, name: str, log_api_exception: bool = True
    ) -> Optional[bool]:
        """Asynchronously switch active input to named input."""
        curr_input_item = await self.__invoke_api_may_need_auth(
            GetCurrentInputCommand(self.device_type),
            log_api_exception=log_api_exception,
        )

        if not curr_input_item:
            _LOGGER.error("Couldn't detect current input")
            return None

        return await self.__invoke_api_may_need_auth(
            ChangeInputCommand(self.device_type, curr_input_item.id, name),
            log_api_exception=log_api_exception,
        )

    async def get_power_state(self, log_api_exception: bool = True) -> Optional[bool]:
        """Asynchronously get device's current power state."""
        item = await self.__invoke_api_may_need_auth(
            GetCurrentPowerStateCommand(self.device_type),
            log_api_exception=log_api_exception,
        )

        if item:
            return bool(item.value)

        return None

    async def pow_on(self, log_api_exception: bool = True) -> Optional[bool]:
        """Asynchronously power device on."""
        return await self.__remote("POW_ON", log_api_exception=log_api_exception)

    async def pow_off(self, log_api_exception: bool = True) -> Optional[bool]:
        """Asynchronously power device off."""
        return await self.__remote("POW_OFF", log_api_exception=log_api_exception)

    async def pow_toggle(self, log_api_exception: bool = True) -> Optional[bool]:
        """Asynchronously toggle device power."""
        return await self.__remote("POW_TOGGLE", log_api_exception=log_api_exception)

    async def vol_up(
        self, num: int = 1, log_api_exception: bool = True
    ) -> Optional[bool]:
        """Asynchronously increase volume by number of steps."""
        return await self.__remote_multiple(
            "VOL_UP", num, log_api_exception=log_api_exception
        )

    async def vol_down(
        self, num: int = 1, log_api_exception: bool = True
    ) -> Optional[bool]:
        """Asynchronously decrease volume by number of steps."""
        return await self.__remote_multiple(
            "VOL_DOWN", num, log_api_exception=log_api_exception
        )

    async def get_current_volume(self, log_api_exception: bool = True) -> Optional[int]:
        """Asynchronously get device's current volume level."""
        return await VizioAsync.get_audio_setting(
            self, "volume", log_api_exception=log_api_exception
        )

    async def is_muted(self, log_api_exception: bool = True) -> Optional[bool]:
        """Asynchronously determine whether or not device is muted."""
        # If None is returned lower() will fail, if not we can do a simple boolean check
        try:
            return (
                await VizioAsync.get_audio_setting(
                    self, "mute", log_api_exception=log_api_exception
                ).lower()
                == "on"
            )
        except AttributeError:
            return None

    def get_max_volume(self) -> int:
        """Get device's max volume based on device type."""
        return MAX_VOLUME[self.device_type]

    async def ch_up(
        self, num: int = 1, log_api_exception: bool = True
    ) -> Optional[bool]:
        """Asynchronously channel up by number of steps."""
        return await self.__remote_multiple(
            "CH_UP", num, log_api_exception=log_api_exception
        )

    async def ch_down(
        self, num: int = 1, log_api_exception: bool = True
    ) -> Optional[bool]:
        """Asynchronously channel down by number of steps."""
        return await self.__remote_multiple(
            "CH_DOWN", num, log_api_exception=log_api_exception
        )

    async def ch_prev(self, log_api_exception: bool = True) -> Optional[bool]:
        """Asynchronously go to previous channel."""
        return await self.__remote("CH_PREV", log_api_exception=log_api_exception)

    async def mute_on(self, log_api_exception: bool = True) -> Optional[bool]:
        """Asynchronously mute sound."""
        return await self.__remote("MUTE_ON", log_api_exception=log_api_exception)

    async def mute_off(self, log_api_exception: bool = True) -> Optional[bool]:
        """Asynchronously unmute sound."""
        return await self.__remote("MUTE_OFF", log_api_exception=log_api_exception)

    async def mute_toggle(self, log_api_exception: bool = True) -> Optional[bool]:
        """Asynchronously toggle sound mute."""
        return await self.__remote("MUTE_TOGGLE", log_api_exception=log_api_exception)

    async def play(self, log_api_exception: bool = True) -> Optional[bool]:
        """Asynchronously emulate 'play' key press."""
        return await self.__remote("PLAY", log_api_exception=log_api_exception)

    async def pause(self, log_api_exception: bool = True) -> Optional[bool]:
        """Asynchronously emulate 'pause' key press."""
        return await self.__remote("PAUSE", log_api_exception=log_api_exception)

    async def remote(self, key: str, log_api_exception: bool = True) -> Optional[bool]:
        """Asynchronously emulate key press by key name."""
        return await self.__remote(key, log_api_exception=log_api_exception)

    def get_remote_keys_list(self) -> List[str]:
        """Get list of remote key names."""
        return KEY_CODE[self.device_type].keys()

    async def get_setting_types_list(
        self, log_api_exception: bool = True
    ) -> Optional[List[str]]:
        """Asynchronously get list of all setting types."""
        items = await self.__invoke_api_may_need_auth(
            GetAllSettingTypesCommand(self.device_type),
            log_api_exception=log_api_exception,
        )

        if items:
            return items

        return None

    async def get_all_settings(
        self, setting_type: str, log_api_exception: bool = True
    ) -> Optional[Dict[str, Union[int, str]]]:
        """Asynchronously get all setting names and corresponding values."""
        item = await self.__invoke_api_may_need_auth(
            GetAllSettingsCommand(self.device_type, setting_type),
            log_api_exception=log_api_exception,
        )

        if item:
            return item

        return None

    async def get_all_settings_options(
        self, setting_type: str, log_api_exception: bool = True
    ) -> Optional[Dict[str, Union[int, str]]]:
        """Asynchronously get all setting names and corresponding options."""
        item = await self.__invoke_api_may_need_auth(
            GetAllSettingsOptionsCommand(self.device_type, setting_type),
            log_api_exception=log_api_exception,
        )

        if item:
            return item

        return None

    async def get_all_settings_options_xlist(
        self, setting_type: str, log_api_exception: bool = True
    ) -> Optional[Dict[str, Union[int, str]]]:
        """Asynchronously get all setting names and corresponding options for settings that are based on a user defined list."""
        item = await self.__invoke_api_may_need_auth(
            GetAllSettingsOptionsXListCommand(self.device_type, setting_type),
            log_api_exception=log_api_exception,
        )

        if item:
            return item

        return None

    async def get_setting(
        self, setting_type: str, setting_name: str, log_api_exception: bool = True
    ) -> Optional[Union[int, str]]:
        """Asynchronously get current value of named setting."""
        item = await self.__invoke_api_may_need_auth(
            GetSettingCommand(self.device_type, setting_type, setting_name),
            log_api_exception=log_api_exception,
        )

        # coerce to int if possible
        if item:
            try:
                return int(item.value)
            except ValueError:
                return item.value

        return None

    async def get_setting_options(
        self, setting_type: str, setting_name: str, log_api_exception: bool = True
    ) -> Optional[Union[int, str]]:
        """Asynchronously get options of named setting."""
        return await self.__invoke_api_may_need_auth(
            GetSettingOptionsCommand(self.device_type, setting_type, setting_name),
            log_api_exception=log_api_exception,
        )

    async def get_setting_options_xlist(
        self, setting_type: str, setting_name: str, log_api_exception: bool = True
    ) -> Optional[Union[int, str]]:
        """Asynchronously get options of named setting for settings based on a user defined list."""
        return await self.__invoke_api_may_need_auth(
            GetSettingOptionsXListCommand(self.device_type, setting_type, setting_name),
            log_api_exception=log_api_exception,
        )

    async def set_setting(
        self,
        setting_type: str,
        setting_name: str,
        new_value: Union[int, str],
        log_api_exception: bool = True,
    ) -> Optional[bool]:
        """Asynchronously set new value for setting."""
        setting_item = await self.__invoke_api_may_need_auth(
            GetSettingCommand(self.device_type, setting_type, setting_name),
            log_api_exception=log_api_exception,
        )

        if not setting_item:
            _LOGGER.error(
                "Couldn't detect setting for %s of setting type %s",
                setting_name,
                setting_type,
            )
            return None

        return await self.__invoke_api_may_need_auth(
            ChangeSettingCommand(
                self.device_type, setting_item.id, setting_type, setting_name, new_value
            ),
            log_api_exception=log_api_exception,
        )

    async def get_all_audio_settings(
        self, log_api_exception: bool = True
    ) -> Optional[Dict[str, Union[int, str]]]:
        """Asynchronously get all audio setting names and corresponding values."""
        return await VizioAsync.get_all_settings(
            self, "audio", log_api_exception=log_api_exception
        )

    async def get_all_audio_settings_options(
        self, log_api_exception: bool = True
    ) -> Optional[Dict[str, Union[int, str]]]:
        """Asynchronously get all audio setting names and corresponding options."""
        return await VizioAsync.get_all_settings_options(
            self, "audio", log_api_exception=log_api_exception
        )

    async def get_audio_setting(
        self, setting_name: str, log_api_exception: bool = True
    ) -> Optional[Union[int, str]]:
        """Asynchronously get current value of named audio setting."""
        return await VizioAsync.get_setting(
            self, "audio", setting_name, log_api_exception=log_api_exception
        )

    async def get_audio_setting_options(
        self, setting_name: str, log_api_exception: bool = True
    ) -> Optional[Union[int, str]]:
        """Asynchronously get options of named audio setting."""
        return await VizioAsync.get_setting_options(
            self, "audio", setting_name, log_api_exception=log_api_exception
        )

    async def set_audio_setting(
        self,
        setting_name: str,
        new_value: Union[int, str],
        log_api_exception: bool = True,
    ) -> Optional[bool]:
        """Asynchronously set new value for named audio setting."""
        return await VizioAsync.set_setting(
            self, "audio", setting_name, new_value, log_api_exception=log_api_exception
        )

    async def get_apps_list(self, country: str = "all") -> List[str]:
        """Get list of known apps by name optionally filtered by supported country."""
        # Assumes "*" means all countries are supported
        apps_list = await gen_apps_list_from_url(session=self._session)
        # Fallback to local list of apps if needed
        if not apps_list:
            apps_list = APPS

        if country.lower() != "all":
            return [
                APP_HOME["name"],
                *sorted(
                    [
                        app["name"]
                        for app in apps_list
                        if "*" in app["country"] or country.lower() in app["country"]
                    ]
                ),
            ]

        return [APP_HOME["name"], *sorted([app["name"] for app in apps_list])]

    async def launch_app(
        self,
        app_name: str,
        apps_list: List[Dict[str, Union[str, List[Union[str, Dict[str, Any]]]]]],
        log_api_exception: bool = True,
    ) -> Optional[bool]:
        """Asynchronously launch known app by name."""
        return await self.__invoke_api_may_need_auth(
            LaunchAppNameCommand(self.device_type, app_name, apps_list),
            log_api_exception=log_api_exception,
        )

    async def launch_app_config(
        self,
        APP_ID: str,
        NAME_SPACE: int,
        MESSAGE: str = None,
        log_api_exception: bool = True,
    ) -> Optional[bool]:
        """Asynchronously launch app using app's config values."""
        return await self.__invoke_api_may_need_auth(
            LaunchAppConfigCommand(self.device_type, APP_ID, NAME_SPACE, MESSAGE),
            log_api_exception=log_api_exception,
        )

    async def get_current_app(
        self,
        apps_list: List[Dict[str, Union[str, List[Union[str, Dict[str, Any]]]]]],
        log_api_exception: bool = True,
    ) -> Optional[str]:
        """Asynchronously get name of currently running app. Returns const APP_NOT_RUNNING if no app is currently running, const UNKNOWN_APP if app config isn't known by pyvizio."""
        return await self.__invoke_api_may_need_auth(
            GetCurrentAppNameCommand(self.device_type, apps_list),
            log_api_exception=log_api_exception,
        )

    async def get_current_app_config(
        self, log_api_exception: bool = True
    ) -> Optional[AppConfig]:
        """Asynchronously get config values of currently running app. Returns empty AppConfig if no app is currently running."""
        return await self.__invoke_api_may_need_auth(
            GetCurrentAppConfigCommand(self.device_type),
            log_api_exception=log_api_exception,
        )


async def async_guess_device_type(
    ip: str, port: Optional[str] = None, timeout: int = DEFAULT_TIMEOUT
) -> str:
    """
    Attempt to guess the device type by getting power state with no auth
    token.

    NOTE:
    The `ip` and `port` values passed in have to be valid for the device in
    order for this to work. This function is being used as part of a zeroconf
    discovery workflow in HomeAssistant which is why it is safe to assume that
    `ip` and `port` are valid.
    """

    if port:
        if ":" in ip:
            raise Exception("Port can't be included in both `ip` and `port` parameters")

        device = VizioAsync(
            "test", ip + ":" + port, "test", "", DEVICE_CLASS_SPEAKER, timeout=timeout
        )
    else:
        if ":" not in ip:
            _LOGGER.warning(
                "May not return correct device type since a port was not specified."
            )
        device = VizioAsync(
            "test", ip, "test", "", DEVICE_CLASS_SPEAKER, timeout=timeout
        )

    if await device.can_connect_with_auth_check():
        return DEVICE_CLASS_SPEAKER
    else:
        return DEVICE_CLASS_TV


class Vizio(VizioAsync):
    """Synchronous class to interact with Vizio SmartCast devices."""

    def __init__(
        self,
        device_id: str,
        ip: str,
        name: str,
        auth_token: str = "",
        device_type: str = DEFAULT_DEVICE_CLASS,
        timeout: int = DEFAULT_TIMEOUT,
    ) -> None:
        """Initialize synchronous class to interact with Vizio SmartCast devices."""
        super(Vizio, self).__init__(
            device_id, ip, name, auth_token, device_type, session=None, timeout=timeout
        )

    @staticmethod
    def discovery_zeroconf(timeout: int = DEFAULT_TIMEOUT) -> List[ZeroconfDevice]:
        """Discover Vizio devices on network using zeroconf."""
        return super(Vizio, Vizio).discovery_zeroconf(timeout)

    @staticmethod
    def discovery_ssdp(timeout: int = DEFAULT_TIMEOUT) -> List[SSDPDevice]:
        """Discover Vizio devices on network using SSDP."""
        return super(Vizio, Vizio).discovery_ssdp(timeout)

    @staticmethod
    @async_to_sync
    async def validate_ha_config(
        ip: str,
        auth_token: str,
        device_type: str,
        session: Optional[ClientSession] = None,
        timeout: int = DEFAULT_TIMEOUT,
    ) -> bool:
        """Return whether or not HomeAssistant config will allow HomeAssistant to make successful calls to Vizio SmartCast API."""
        return await super(Vizio, Vizio).validate_ha_config(
            ip, auth_token, device_type, session=session, timeout=timeout
        )

    @staticmethod
    @async_to_sync
    async def get_unique_id(
        ip: str, device_type: str, timeout: int = DEFAULT_TIMEOUT
    ) -> Optional[str]:
        """Get unique identifier for Vizio device."""
        return await super(Vizio, Vizio).get_unique_id(ip, device_type, timeout=timeout)

    @async_to_sync
    async def can_connect_with_auth_check(self) -> bool:
        """Return whether or not device API can be connected to with valid authorization."""
        return await super(Vizio, self).can_connect_with_auth_check()

    @async_to_sync
    async def can_connect_no_auth_check(self) -> bool:
        """Return whether or not device API can be connected to regardless of auth config."""
        return await super(Vizio, self).can_connect_no_auth_check()

    @async_to_sync
    async def get_esn(self, log_api_exception: bool = True) -> Optional[str]:
        """Get device's ESN (electronic serial number?)."""
        return await super(Vizio, self).get_esn(log_api_exception=log_api_exception)

    @async_to_sync
    async def get_serial_number(self, log_api_exception: bool = True) -> Optional[str]:
        """Get device's serial number."""
        return await super(Vizio, self).get_serial_number(
            log_api_exception=log_api_exception
        )

    @async_to_sync
    async def get_version(self, log_api_exception: bool = True) -> Optional[str]:
        """Get SmartCast software version on device."""
        return await super(Vizio, self).get_version(log_api_exception=log_api_exception)

    @async_to_sync
    async def get_model_name(self, log_api_exception: bool = True) -> Optional[str]:
        """Get device's model number."""
        return await super(Vizio, self).get_model_name(
            log_api_exception=log_api_exception
        )

    @async_to_sync
    async def start_pair(
        self, log_api_exception: bool = True
    ) -> Optional[BeginPairResponse]:
        """Begin pairing process to obtain challenge type and challenge token."""
        return await super(Vizio, self).start_pair(log_api_exception=log_api_exception)

    @async_to_sync
    async def stop_pair(self, log_api_exception: bool = True) -> Optional[bool]:
        """Cancel pairing process."""
        return await super(Vizio, self).stop_pair(log_api_exception=log_api_exception)

    @async_to_sync
    async def pair(
        self, ch_type: int, token: int, pin: str, log_api_exception: bool = True
    ) -> Optional[PairChallengeResponse]:
        """Complete pairing process to obtain auth token."""
        return await super(Vizio, self).pair(
            ch_type, token, pin, log_api_exception=log_api_exception
        )

    @async_to_sync
    async def get_inputs_list(
        self, log_api_exception: bool = True
    ) -> Optional[List[InputItem]]:
        """Get list of available inputs."""
        return await super(Vizio, self).get_inputs_list(
            log_api_exception=log_api_exception
        )

    @async_to_sync
    async def get_current_input(self, log_api_exception: bool = True) -> Optional[str]:
        """Get device's active input."""
        return await super(Vizio, self).get_current_input(
            log_api_exception=log_api_exception
        )

    @async_to_sync
    async def next_input(self, log_api_exception: bool = True) -> Optional[bool]:
        """Switch active input to next input."""
        return await super(Vizio, self).next_input(log_api_exception=log_api_exception)

    @async_to_sync
    async def set_input(
        self, name: str, log_api_exception: bool = True
    ) -> Optional[bool]:
        """Switch active input to named input."""
        return await super(Vizio, self).set_input(
            name, log_api_exception=log_api_exception
        )

    @async_to_sync
    async def get_power_state(self, log_api_exception: bool = True) -> Optional[bool]:
        """Get device's current power state."""
        return await super(Vizio, self).get_power_state(
            log_api_exception=log_api_exception
        )

    @async_to_sync
    async def pow_on(self, log_api_exception: bool = True) -> Optional[bool]:
        """Power device off."""
        return await super(Vizio, self).pow_on(log_api_exception=log_api_exception)

    @async_to_sync
    async def pow_off(self, log_api_exception: bool = True) -> Optional[bool]:
        """Power device off."""
        return await super(Vizio, self).pow_off(log_api_exception=log_api_exception)

    @async_to_sync
    async def pow_toggle(self, log_api_exception: bool = True) -> Optional[bool]:
        """Toggle device power."""
        return await super(Vizio, self).pow_toggle(log_api_exception=log_api_exception)

    @async_to_sync
    async def vol_up(
        self, num: int = 1, log_api_exception: bool = True
    ) -> Optional[bool]:
        """Increase volume by number of steps."""
        return await super(Vizio, self).vol_up(num, log_api_exception=log_api_exception)

    @async_to_sync
    async def vol_down(
        self, num: int = 1, log_api_exception: bool = True
    ) -> Optional[bool]:
        """Decrease volume by number of steps."""
        return await super(Vizio, self).vol_down(
            num, log_api_exception=log_api_exception
        )

    @async_to_sync
    async def get_current_volume(self, log_api_exception: bool = True) -> Optional[int]:
        """Get device's current volume level."""
        return await super(Vizio, self).get_current_volume(
            log_api_exception=log_api_exception
        )

    @async_to_sync
    async def is_muted(self, log_api_exception: bool = True) -> Optional[bool]:
        """Return whether or not device is muted."""
        return await super(Vizio, self).is_muted(log_api_exception=log_api_exception)

    def get_max_volume(self) -> int:
        """Return device's max volume based on device type."""
        return super(Vizio, self).get_max_volume()

    @async_to_sync
    async def ch_up(
        self, num: int = 1, log_api_exception: bool = True
    ) -> Optional[bool]:
        """Channel up by number of steps."""
        return await super(Vizio, self).ch_up(num, log_api_exception=log_api_exception)

    @async_to_sync
    async def ch_down(
        self, num: int = 1, log_api_exception: bool = True
    ) -> Optional[bool]:
        """Channel down by number of steps."""
        return await super(Vizio, self).ch_down(
            num, log_api_exception=log_api_exception
        )

    @async_to_sync
    async def ch_prev(self, log_api_exception: bool = True) -> Optional[bool]:
        """Go to previous channel."""
        return await super(Vizio, self).ch_prev(log_api_exception=log_api_exception)

    @async_to_sync
    async def mute_on(self, log_api_exception: bool = True) -> Optional[bool]:
        """Mute sound."""
        return await super(Vizio, self).mute_on(log_api_exception=log_api_exception)

    @async_to_sync
    async def mute_off(self, log_api_exception: bool = True) -> Optional[bool]:
        """Unmute sound."""
        return await super(Vizio, self).mute_off(log_api_exception=log_api_exception)

    @async_to_sync
    async def mute_toggle(self, log_api_exception: bool = True) -> Optional[bool]:
        """Toggle sound mute."""
        return await super(Vizio, self).mute_toggle(log_api_exception=log_api_exception)

    @async_to_sync
    async def play(self, log_api_exception: bool = True) -> Optional[bool]:
        """Emulate 'play' key press."""
        return await super(Vizio, self).play(log_api_exception=log_api_exception)

    @async_to_sync
    async def pause(self, log_api_exception: bool = True) -> Optional[bool]:
        """Emulate 'pause' key press."""
        return await super(Vizio, self).pause(log_api_exception=log_api_exception)

    @async_to_sync
    async def remote(self, key: str, log_api_exception: bool = True) -> Optional[bool]:
        """Emulate key press by key name."""
        return await super(Vizio, self).remote(key, log_api_exception=log_api_exception)

    def get_remote_keys_list(self) -> List[str]:
        """Get list of remote key names."""
        return super(Vizio, self).get_remote_keys_list()

    @async_to_sync
    async def get_setting_types_list(
        self, setting_type: str, log_api_exception: bool = True
    ) -> Optional[List[str]]:
        """Get list of all setting types."""
        return await super(Vizio, self).get_setting_types_list(
            log_api_exception=log_api_exception
        )

    @async_to_sync
    async def get_all_settings(
        self, setting_type: str, log_api_exception: bool = True
    ) -> Optional[Dict[str, Union[int, str]]]:
        """Get all setting names and corresponding values."""
        return await super(Vizio, self).get_all_settings(
            setting_type, log_api_exception=log_api_exception
        )

    @async_to_sync
    async def get_all_settings_options(
        self, setting_type: str, log_api_exception: bool = True
    ) -> Optional[Dict[str, Union[int, str]]]:
        """Get all setting names and corresponding options."""
        return await super(Vizio, self).get_all_settings_options(
            setting_type, log_api_exception=log_api_exception
        )

    @async_to_sync
    async def get_setting(
        self, setting_type: str, setting_name: str, log_api_exception: bool = True
    ) -> Optional[Union[int, str]]:
        """Get current value of named setting."""
        return await super(Vizio, self).get_setting(
            setting_type, setting_name, log_api_exception=log_api_exception
        )

    @async_to_sync
    async def get_setting_options(
        self, setting_type: str, setting_name: str, log_api_exception: bool = True
    ) -> Optional[Union[int, str]]:
        """Get options of named setting."""
        return await super(Vizio, self).get_setting_options(
            setting_type, setting_name, log_api_exception=log_api_exception
        )

    @async_to_sync
    async def set_setting(
        self,
        setting_type: str,
        setting_name: str,
        new_value: Union[int, str],
        log_api_exception: bool = True,
    ) -> Optional[bool]:
        """Set new value for named setting."""
        return await super(Vizio, self).set_setting(
            setting_type, setting_name, new_value, log_api_exception=log_api_exception
        )

    @async_to_sync
    async def get_all_audio_settings(
        self, log_api_exception: bool = True
    ) -> Optional[Dict[str, Union[int, str]]]:
        """Get all audio setting names and corresponding values."""
        return await super(Vizio, self).get_all_audio_settings(
            log_api_exception=log_api_exception
        )

    @async_to_sync
    async def get_all_audio_settings_options(
        self, log_api_exception: bool = True
    ) -> Optional[Dict[str, Union[int, str]]]:
        """Get all audio setting names and corresponding options."""
        return await super(Vizio, self).get_all_audio_settings_options(
            log_api_exception=log_api_exception
        )

    @async_to_sync
    async def get_audio_setting(
        self, setting_name: str, log_api_exception: bool = True
    ) -> Optional[Union[int, str]]:
        """Get current value of named audio setting."""
        return await super(Vizio, self).get_audio_setting(
            setting_name, log_api_exception=log_api_exception
        )

    @async_to_sync
    async def get_audio_setting_options(
        self, setting_name: str, log_api_exception: bool = True
    ) -> Optional[Union[int, str]]:
        """Get options of named audio setting."""
        return await super(Vizio, self).get_audio_setting_options(
            setting_name, log_api_exception=log_api_exception
        )

    @async_to_sync
    async def set_audio_setting(
        self,
        setting_name: str,
        new_value: Union[int, str],
        log_api_exception: bool = True,
    ) -> Optional[bool]:
        """Set new value for named audio setting."""
        return await super(Vizio, self).set_audio_setting(
            setting_name, new_value, log_api_exception=log_api_exception
        )

    @async_to_sync
    async def get_apps_list(self, country: str = None) -> List[str]:
        """Get list of known apps by name optionally filtered by supported country."""
        return await super(Vizio, self).get_apps_list(country=country)

    @async_to_sync
    async def launch_app(
        self,
        app_name: str,
        apps_list: List[Dict[str, Union[str, List[Union[str, Dict[str, Any]]]]]],
        log_api_exception: bool = True,
    ) -> Optional[bool]:
        """Launch known app by name."""
        return await super(Vizio, self).launch_app(
            app_name, apps_list, log_api_exception=log_api_exception
        )

    @async_to_sync
    async def launch_app_config(
        self,
        app_name: str,
        APP_ID: str,
        NAME_SPACE: int,
        MESSAGE: str = None,
        log_api_exception: bool = True,
    ) -> Optional[bool]:
        """Launch app using app's config values."""
        return await super(Vizio, self).launch_app_config(
            APP_ID, NAME_SPACE, MESSAGE, log_api_exception=log_api_exception
        )

    @async_to_sync
    async def get_current_app(
        self,
        apps_list: List[Dict[str, Union[str, List[Union[str, Dict[str, Any]]]]]],
        log_api_exception: bool = True,
    ) -> Optional[str]:
        """Get name of currently running app. Returns const APP_NOT_RUNNING if no app is currently running, const UNKNOWN_APP if app config isn't known by pyvizio."""
        return await super(Vizio, self).get_current_app(
            apps_list, log_api_exception=log_api_exception
        )

    @async_to_sync
    async def get_current_app_config(
        self, log_api_exception: bool = True
    ) -> Optional[AppConfig]:
        """Get config values of currently running app. Returns empty AppConfig if no app is currently running."""
        return await super(Vizio, self).get_current_app_config(
            log_api_exception=log_api_exception
        )


@async_to_sync
async def guess_device_type(
    ip: str, port: Optional[str] = None, timeout: int = DEFAULT_TIMEOUT
) -> str:
    """
    Attempt to guess the device type by getting power state with no auth
    token.

    NOTE:
    The `ip` and `port` values passed in have to be valid for the device in
    order for this to work. This function is being used as part of a zeroconf
    discovery workflow in HomeAssistant which is why it is safe to assume that
    `ip` and `port` are valid.
    """

    return await async_guess_device_type(ip, port, timeout)



"""pyvizio constants."""

DEVICE_CLASS_SPEAKER = "speaker"
DEVICE_CLASS_TV = "tv"

DEFAULT_DEVICE_ID = "pyvizio"
DEFAULT_DEVICE_CLASS = DEVICE_CLASS_TV
DEFAULT_DEVICE_NAME = "Python Vizio"
DEFAULT_PORTS = [7345, 9000]
DEFAULT_TIMEOUT = 5

MAX_VOLUME = {DEVICE_CLASS_TV: 100, DEVICE_CLASS_SPEAKER: 31}

# Current Input when app is active
INPUT_APPS = ["SMARTCAST", "CAST"]

# App name returned when it is not in app dictionary
UNKNOWN_APP = "_UNKNOWN_APP"
NO_APP_RUNNING = "_NO_APP_RUNNING"
SMARTCAST_HOME = "SmartCast Home"

APP_CAST = "Cast"

# NAME_SPACE values that appear to be equivalent
EQUIVALENT_NAME_SPACES = (2, 4)

APP_HOME = {
    "name": SMARTCAST_HOME,
    "country": ["*"],
    "config": [
        {
            "NAME_SPACE": 4,
            "APP_ID": "1",
            "MESSAGE": "http://127.0.0.1:12345/scfs/sctv/main.html",
        }
    ],
}

# No longer needed but kept around in case the external source for APPS is unavailable
APPS = [
    {
        "name": "Prime Video",
        "country": ["*"],
        "id": ["33"],
        "config": [
            {
                "APP_ID": "4",
                "NAME_SPACE": 4,
                "MESSAGE": "https://atv-ext.amazon.com/blast-app-hosting/html5/index.html?deviceTypeID=A3OI4IHTNZQWDD",
            },
            {"NAME_SPACE": 2, "APP_ID": "4", "MESSAGE": "None"},
        ],
    },
    {
        "name": "CBS All Access",
        "country": ["usa"],
        "id": ["9"],
        "config": [{"NAME_SPACE": 2, "APP_ID": "37", "MESSAGE": "None"}],
    },
    {
        "name": "CBS News",
        "country": ["usa", "can"],
        "id": ["56"],
        "config": [{"NAME_SPACE": 2, "APP_ID": "42", "MESSAGE": "None"}],
    },
    {
        "name": "Crackle",
        "country": ["usa"],
        "id": ["8"],
        "config": [{"NAME_SPACE": 2, "APP_ID": "5", "MESSAGE": "None"}],
    },
    {
        "name": "Curiosity Stream",
        "country": ["usa", "can"],
        "id": ["37"],
        "config": [{"NAME_SPACE": 2, "APP_ID": "12", "MESSAGE": "None"}],
    },
    {
        "name": "Fandango Now",
        "country": ["usa"],
        "id": ["24"],
        "config": [{"NAME_SPACE": 2, "APP_ID": "7", "MESSAGE": "None"}],
    },
    {
        "name": "FilmRise",
        "country": ["usa"],
        "id": ["47"],
        "config": [{"NAME_SPACE": 2, "APP_ID": "24", "MESSAGE": "None"}],
    },
    {
        "name": "Flixfling",
        "country": ["*"],
        "id": ["49"],
        "config": [{"NAME_SPACE": 2, "APP_ID": "36", "MESSAGE": "None"}],
    },
    {
        "name": "Haystack TV",
        "country": ["usa", "can"],
        "id": ["35"],
        "config": [
            {
                "NAME_SPACE": 0,
                "APP_ID": "898AF734",
                "MESSAGE": '{"CAST_NAMESPACE":"urn:x-cast:com.google.cast.media","CAST_MESSAGE":{"type":"LOAD","media":{},"autoplay":true,"currentTime":0,"customData":{"platform":"sctv"}}}',
            }
        ],
    },
    {
        "name": "Hulu",
        "country": ["usa"],
        "id": ["19"],
        "config": [
            {
                "APP_ID": "3",
                "NAME_SPACE": 4,
                "MESSAGE": "https://viziosmartcast.app.hulu.com/livingroom/viziosmartcast/1/index.html#initialize",
            },
            {"NAME_SPACE": 2, "APP_ID": "3", "MESSAGE": "None"},
        ],
    },
    {
        "name": "iHeartRadio",
        "country": ["usa"],
        "id": ["11"],
        "config": [{"NAME_SPACE": 2, "APP_ID": "6", "MESSAGE": "None"}],
    },
    {
        "name": "NBC",
        "country": ["usa"],
        "id": ["43"],
        "config": [{"NAME_SPACE": 2, "APP_ID": "10", "MESSAGE": "None"}],
    },
    {
        "name": "Netflix",
        "country": ["*"],
        "id": ["34"],
        "config": [{"NAME_SPACE": 3, "APP_ID": "1", "MESSAGE": "None"}],
    },
    {
        "name": "Plex",
        "country": ["usa", "can"],
        "id": ["40"],
        "config": [
            {
                "APP_ID": "9",
                "NAME_SPACE": 4,
                "MESSAGE": "https://plex.tv/web/tv/vizio-smartcast",
            },
            {"NAME_SPACE": 2, "APP_ID": "9", "MESSAGE": "None"},
        ],
    },
    {
        "name": "Pluto TV",
        "country": ["usa"],
        "id": ["12"],
        "config": [
            {"APP_ID": "65", "NAME_SPACE": 4, "MESSAGE": "https://smartcast.pluto.tv"},
            {
                "NAME_SPACE": 0,
                "APP_ID": "E6F74C01",
                "MESSAGE": '{"CAST_NAMESPACE":"urn:x-cast:tv.pluto","CAST_MESSAGE":{"command":"initializePlayback","channel":"","episode":"","time":0}}',
            },
        ],
    },
    {
        "name": "RedBox",
        "country": ["usa"],
        "id": ["55"],
        "config": [{"NAME_SPACE": 2, "APP_ID": "41", "MESSAGE": "None"}],
    },
    {
        "name": "TasteIt",
        "country": ["*"],
        "id": ["52"],
        "config": [{"NAME_SPACE": 2, "APP_ID": "26", "MESSAGE": "None"}],
    },
    {
        "name": "Toon Goggles",
        "country": ["usa", "can"],
        "id": ["46"],
        "config": [{"NAME_SPACE": 2, "APP_ID": "21", "MESSAGE": "None"}],
    },
    {
        "name": "Vudu",
        "country": ["usa"],
        "id": ["6"],
        "config": [
            {
                "APP_ID": "31",
                "NAME_SPACE": 4,
                "MESSAGE": "https://my.vudu.com/castReceiver/index.html?launch-source=app-icon",
            }
        ],
    },
    {
        "name": "XUMO",
        "country": ["usa"],
        "id": ["27"],
        "config": [
            {
                "NAME_SPACE": 0,
                "APP_ID": "36E1EA1F",
                "MESSAGE": '{"CAST_NAMESPACE":"urn:x-cast:com.google.cast.media","CAST_MESSAGE":{"type":"LOAD","media":{},"autoplay":true,"currentTime":0,"customData":{}}}',
            }
        ],
    },
    {
        "name": "YouTubeTV",
        "country": ["usa", "mexico"],
        "id": ["45"],
        "config": [{"NAME_SPACE": 5, "APP_ID": "3", "MESSAGE": "None"}],
    },
    {
        "name": "YouTube",
        "country": ["*"],
        "id": ["44"],
        "config": [{"NAME_SPACE": 5, "APP_ID": "1", "MESSAGE": "None"}],
    },
    {
        "name": "Baeble",
        "country": ["usa"],
        "id": ["39"],
        "config": [{"NAME_SPACE": 2, "APP_ID": "11", "MESSAGE": "None"}],
    },
    {
        "name": "DAZN",
        "country": ["usa", "can"],
        "id": ["57"],
        "config": [{"NAME_SPACE": 2, "APP_ID": "34", "MESSAGE": "None"}],
    },
    {
        "name": "FitFusion by Jillian Michaels",
        "country": ["usa", "can"],
        "id": ["54"],
        "config": [{"NAME_SPACE": 2, "APP_ID": "39", "MESSAGE": "None"}],
    },
    {
        "name": "Newsy",
        "country": ["usa", "can"],
        "id": ["38"],
        "config": [{"NAME_SPACE": 2, "APP_ID": "15", "MESSAGE": "None"}],
    },
    {
        "name": "Cocoro TV",
        "country": ["usa", "can"],
        "id": ["63"],
        "config": [{"NAME_SPACE": 2, "APP_ID": "55", "MESSAGE": "None"}],
    },
    {
        "name": "ConTV",
        "country": ["usa", "can"],
        "id": ["41"],
        "config": [{"NAME_SPACE": 2, "APP_ID": "18", "MESSAGE": "None"}],
    },
    {
        "name": "Dove Channel",
        "country": ["usa", "can"],
        "id": ["42"],
        "config": [{"NAME_SPACE": 2, "APP_ID": "16", "MESSAGE": "None"}],
    },
    {
        "name": "Love Destination",
        "country": ["*"],
        "id": ["64"],
        "config": [{"NAME_SPACE": 2, "APP_ID": "57", "MESSAGE": "None"}],
    },
    {
        "name": "WatchFree",
        "country": ["usa"],
        "id": ["48"],
        "config": [{"NAME_SPACE": 2, "APP_ID": "22", "MESSAGE": "None"}],
    },
    {
        "name": "AsianCrush",
        "country": ["usa", "can"],
        "id": ["50"],
        "config": [
            {
                "NAME_SPACE": 2,
                "APP_ID": "27",
                "MESSAGE": "https://html5.asiancrush.com/?ua=viziosmartcast",
            }
        ],
    },
    {
        "name": "Disney+",
        "country": ["usa"],
        "id": ["51"],
        "config": [
            {
                "NAME_SPACE": 4,
                "APP_ID": "75",
                "MESSAGE": "https://cd-dmgz.bamgrid.com/bbd/vizio_tv/index.html",
            }
        ],
    },
]



"""Vizio SmartCast API commands and class for individual item settings."""

from typing import Any, Dict, Union

from pyvizio.api._protocol import (
    ACTION_MODIFY,
    ENDPOINT,
    ITEM_CNAME,
    PATH_MODEL,
    ResponseKey,
)
from pyvizio.api.base import CommandBase, InfoCommandBase
from pyvizio.helpers import dict_get_case_insensitive, get_value_from_path


class GetDeviceInfoCommand(InfoCommandBase):
    """Command to get device info."""

    def __init__(self, device_type: str) -> None:
        """Initialize command to get device info."""
        super(GetDeviceInfoCommand, self).__init__(ENDPOINT[device_type]["DEVICE_INFO"])
        self.paths = PATH_MODEL[device_type]

    def process_response(self, json_obj: Dict[str, Any]) -> bool:
        """Return response to command to get device info."""
        return dict_get_case_insensitive(json_obj, ResponseKey.ITEMS, [{}])[0]


class GetModelNameCommand(GetDeviceInfoCommand):
    """Command to get device model name."""

    def __init__(self, device_type: str) -> None:
        """Initialize command to get device model name."""
        super(GetModelNameCommand, self).__init__(device_type)

    def process_response(self, json_obj: Dict[str, Any]) -> bool:
        """Return response to command to get device model name."""
        return get_value_from_path(
            dict_get_case_insensitive(
                super(GetModelNameCommand, self).process_response(json_obj),
                ResponseKey.VALUE,
                {},
            ),
            self.paths,
        )


class Item(object):
    """Individual item setting."""

    def __init__(self, json_obj: Dict[str, Any]) -> None:
        """Initialize individual item setting."""
        self.id = None
        id = dict_get_case_insensitive(json_obj, ResponseKey.HASHVAL)
        if id is not None:
            self.id = int(id)

        self.c_name = dict_get_case_insensitive(json_obj, ResponseKey.CNAME)
        self.type = dict_get_case_insensitive(json_obj, ResponseKey.TYPE)
        self.name = dict_get_case_insensitive(json_obj, ResponseKey.NAME)
        self.value = dict_get_case_insensitive(json_obj, ResponseKey.VALUE)

        self.min = None
        min = dict_get_case_insensitive(json_obj, ResponseKey.MINIMUM)
        if min is not None:
            self.min = int(min)

        self.max = None
        max = dict_get_case_insensitive(json_obj, ResponseKey.MAXIMUM)
        if max is not None:
            self.max = int(max)

        self.center = None
        center = dict_get_case_insensitive(json_obj, ResponseKey.CENTER)
        if center is not None:
            self.center = int(center)

        self.choices = dict_get_case_insensitive(json_obj, ResponseKey.ELEMENTS, [])

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.__dict__})"

    def __eq__(self, other) -> bool:
        return self is other or (
            self.c_name == other.c_name
            and self.type == other.type
            and self.name == other.name
            and self.value == other.value
        )


class DefaultReturnItem(object):
    """Mock individual item setting response when item is not found."""

    def __init__(self, value: Any) -> None:
        """Initialize mock individual item setting response when item is not found."""
        self.value = value

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.__dict__})"

    def __eq__(self, other) -> bool:
        return self is other or self.__dict__ == other.__dict__


class ItemInfoCommandBase(InfoCommandBase):
    """Command to get individual item setting."""

    def __init__(
        self, device_type: str, item_name: str, default_return: Union[int, str] = None
    ) -> None:
        """Initialize command to get individual item setting."""
        super(ItemInfoCommandBase, self).__init__(ENDPOINT[device_type][item_name])
        self.item_name = item_name.upper()
        self.default_return = default_return

    def process_response(self, json_obj: Dict[str, Any]) -> Any:
        """Return response to command to get individual item setting."""
        items = [
            Item(item)
            for item in dict_get_case_insensitive(json_obj, ResponseKey.ITEMS, [])
        ]

        for itm in items:
            if itm.c_name.lower() in (
                ITEM_CNAME.get(self.item_name, ""),
                self.item_name,
            ) and (
                itm.value is not None
                or itm.center is not None
                or itm.choices is not None
            ):
                return itm

        if self.default_return is not None:
            return DefaultReturnItem(self.default_return)

        return None


class ItemCommandBase(CommandBase):
    """Command to set value of individual item setting."""

    def __init__(
        self, device_type: str, item_name: str, id: int, value: Union[int, str]
    ) -> None:
        """Initialize command to set value of individual item setting."""
        super(ItemCommandBase, self).__init__(ENDPOINT[device_type][item_name])
        self.item_name = item_name

        self.VALUE = value
        # noinspection SpellCheckingInspection
        self.HASHVAL = int(id)
        self.REQUEST = ACTION_MODIFY.upper()


class GetCurrentPowerStateCommand(ItemInfoCommandBase):
    """Command to get current power state of device."""

    def __init__(self, device_type: str) -> None:
        """Initialize command to get current power state of device."""
        super(GetCurrentPowerStateCommand, self).__init__(device_type, "POWER_MODE", 0)


class GetESNCommand(ItemInfoCommandBase):
    """Command to get device ESN (electronic serial number?)."""

    def __init__(self, device_type: str) -> None:
        """Initialize command to get device ESN (electronic serial number?)."""
        super(GetESNCommand, self).__init__(device_type, "ESN")


class GetSerialNumberCommand(ItemInfoCommandBase):
    """Command to get device serial number."""

    def __init__(self, device_type: str) -> None:
        """Initialize command to get device serial number."""
        super(GetSerialNumberCommand, self).__init__(device_type, "SERIAL_NUMBER")


class GetVersionCommand(ItemInfoCommandBase):
    """Command to get SmartCast software version."""

    def __init__(self, device_type: str) -> None:
        """Initialize command to get SmartCast software version."""
        super(GetVersionCommand, self).__init__(device_type, "VERSION")



"""pyvizio utility module."""

import json
from typing import Any, Dict, List, Union

from aiohttp import ClientError, ClientSession

from pyvizio.util.const import (
    APK_SOURCE_PATH,
    APP_NAMES_FILE,
    APP_NAMES_URL,
    APP_PAYLOADS_FILE,
    APP_PAYLOADS_URL,
    RESOURCE_PATH,
)


async def gen_apps_list_from_url(
    app_names_url: str = APP_NAMES_URL,
    app_payloads_url: str = APP_PAYLOADS_URL,
    session: ClientSession = None,
) -> Optional[List[Dict[str, Union[str, List[Union[str, Dict[str, Any]]]]]]]:
    """Get app JSON files from external URLs and return list of apps for use in pyvizio."""
    headers = {"Content-Type": "application/json"}
    try:
        if session:
            response = await session.get(
                app_names_url, headers=headers, raise_for_status=True
            )
            app_names = await response.json(content_type=None)
            response = await session.get(
                app_payloads_url, headers=headers, raise_for_status=True
            )
            app_configs = await response.json(content_type=None)
        else:
            async with ClientSession() as local_session:
                response = await local_session.get(
                    app_names_url, headers=headers, raise_for_status=True
                )
                app_names = await response.json(content_type=None)
                response = await local_session.get(
                    app_payloads_url, headers=headers, raise_for_status=True
                )
                app_configs = await response.json(content_type=None)

        return gen_apps_list(app_names, app_configs)
    except ClientError:
        return None


def gen_apps_list_from_src(
    apk_source_path: str = APK_SOURCE_PATH, resource_path: str = RESOURCE_PATH
) -> List[Dict[str, Union[str, List[Union[str, Dict[str, Any]]]]]]:
    """Parse JSON from VizioCast Android app source in `apk_source_path`/`resource_path` and return list of apps for use in pyvizio."""
    base_path = f"{apk_source_path}/{resource_path}"
    app_names_filepath = f"{base_path}/{APP_NAMES_FILE}"
    app_configs_filepath = f"{base_path}/{APP_PAYLOADS_FILE}"

    with open(app_names_filepath) as f:
        app_names = json.load(f)

    with open(app_configs_filepath) as f:
        app_configs = json.load(f)

    return gen_apps_list(app_names, app_configs)


def gen_apps_list(
    app_names: List[Dict[str, Any]], app_configs: List[Dict[str, Any]]
) -> List[Dict[str, Union[str, List[Union[str, Dict[str, Any]]]]]]:
    """Parse list of app names and app configs and return list of apps for use in pyvizio."""
    pyvizio_apps: List[Dict[str, Union[str, List[Union[str, Dict[str, Any]]]]]] = []

    for app_name in app_names:
        # returns first app where condition is true
        app_config = next(
            (
                app_config
                for app_config in app_configs
                if app_config["id"] == app_name["id"]
            )
        )

        if app_config:
            config_json = app_config["chipsets"]["*"][0]["app_type_payload"]
            config = json.loads(config_json)
            app_already_exists = False
            for pyvizio_app in pyvizio_apps:
                if pyvizio_app["name"].lower() == app_name["name"].lower():
                    pyvizio_app["id"].append(app_name["id"])
                    pyvizio_app["config"].append(config)
                    app_already_exists = True
                    break

            if not app_already_exists:
                pyvizio_apps.append(
                    {
                        "name": app_name["name"],
                        "country": [country.lower() for country in app_name["country"]],
                        "id": [app_name["id"]],
                        "config": [config],
                    }
                )

    return pyvizio_apps



"""Vizio SmartCast API commands for audio settings."""

from typing import Any, Dict, List, Optional, Union

from pyvizio.api._protocol import (
    ENDPOINT,
    TYPE_LIST,
    TYPE_MENU,
    TYPE_SLIDER,
    TYPE_VALUE,
    TYPE_X_LIST,
    ResponseKey,
)
from pyvizio.api.item import Item, ItemCommandBase, ItemInfoCommandBase
from pyvizio.helpers import dict_get_case_insensitive


class GetAllSettingTypesCommand(ItemInfoCommandBase):
    """Command to get list of all setting types."""

    def __init__(self, device_type: str) -> None:
        """Initialize command to get list of all setting types."""
        super(GetAllSettingTypesCommand, self).__init__(device_type, "SETTINGS")
        ItemInfoCommandBase.url.fset(self, f"{ENDPOINT[device_type]['SETTINGS']}")

    def process_response(self, json_obj: Dict[str, Any]) -> List[str]:
        """Return response to command to get list of all setting types."""
        items = [
            Item(item)
            for item in dict_get_case_insensitive(json_obj, ResponseKey.ITEMS, [])
        ]
        return [
            item.c_name
            for item in items
            if item.type.lower() == TYPE_MENU
            and item.c_name not in ("cast", "input", "devices", "network")
        ]


class GetAllSettingsCommand(ItemInfoCommandBase):
    """Command to get list of all setting names and corresponding values."""

    def __init__(self, device_type: str, setting_type: str) -> None:
        """Initialize command to get list of all setting names and corresponding values."""
        super(GetAllSettingsCommand, self).__init__(device_type, "SETTINGS")
        self.setting_type = setting_type.lower()
        ItemInfoCommandBase.url.fset(
            self, f"{ENDPOINT[device_type]['SETTINGS']}/{setting_type}"
        )

    def process_response(self, json_obj: Dict[str, Any]) -> Dict[str, Union[int, str]]:
        """Return response to command to get list of all setting names and corresponding values."""
        items = [
            Item(item)
            for item in dict_get_case_insensitive(json_obj, ResponseKey.ITEMS, [])
        ]
        return {
            item.c_name: item.value
            for item in items
            if item.type.lower() in (TYPE_LIST, TYPE_SLIDER, TYPE_VALUE)
        }


class GetSettingCommand(ItemInfoCommandBase):
    """Command to get value of a setting by name."""

    def __init__(self, device_type: str, setting_type: str, setting_name: str) -> None:
        """Initialize command to get value of a setting by name."""
        super(GetSettingCommand, self).__init__(device_type, "SETTINGS", 0)
        self.item_name = setting_name.lower()
        self.setting_type = setting_type.lower()
        ItemInfoCommandBase.url.fset(
            self, f"{ENDPOINT[device_type]['SETTINGS']}/{setting_type}/{setting_name}"
        )


class GetAllSettingsOptionsCommand(ItemInfoCommandBase):
    """Command to get list of all setting names and corresponding options."""

    def __init__(self, device_type: str, setting_type: str) -> None:
        """Initialize command to get list of all setting names and corresponding options."""
        super(GetAllSettingsOptionsCommand, self).__init__(
            device_type, "SETTINGS_OPTIONS"
        )
        self.setting_type = setting_type.lower()
        ItemInfoCommandBase.url.fset(
            self, f"{ENDPOINT[device_type]['SETTINGS_OPTIONS']}/{setting_type}"
        )

    def process_response(
        self, json_obj: Dict[str, Any]
    ) -> Dict[str, Union[List[str], Dict[str, Union[int, str]]]]:
        """Return response to command to get list of all setting names and corresponding options."""
        items = [
            Item(item)
            for item in dict_get_case_insensitive(json_obj, ResponseKey.ITEMS, [])
        ]

        settings_options = {}
        for item in items:
            if item.type.lower() in (TYPE_SLIDER, TYPE_VALUE):
                settings_options[item.c_name] = {"min": item.min, "max": item.max}
                if item.center is not None:
                    settings_options[item.c_name].update({"default": item.center})
            elif item.type.lower() == TYPE_LIST:
                settings_options[item.c_name] = item.choices.copy()

        return settings_options


class GetSettingOptionsCommand(GetAllSettingsOptionsCommand):
    """Command to get options of a setting by name."""

    def __init__(self, device_type: str, setting_type: str, setting_name: str) -> None:
        """Initialize command to get options of a setting by name."""
        self.setting_name = setting_name
        self.setting_type = setting_type.lower()
        super(GetSettingOptionsCommand, self).__init__(device_type, self.setting_type)

    def process_response(
        self, json_obj: Dict[str, Any]
    ) -> Optional[Union[List[str], Dict[str, Union[int, str]]]]:
        """Return response to command to get options of a setting by name."""
        return (
            super(GetSettingOptionsCommand, self)
            .process_response(json_obj)
            .get(self.setting_name)
        )


class GetAllSettingsOptionsXListCommand(ItemInfoCommandBase):
    """Command to get list of all setting names and corresponding options for settings of type XList."""

    def __init__(self, device_type: str, setting_type: str) -> None:
        """Initialize command to get list of all setting names and corresponding options for settings of type XList."""
        super(GetAllSettingsOptionsXListCommand, self).__init__(device_type, "SETTINGS")
        self.setting_type = setting_type.lower()
        ItemInfoCommandBase.url.fset(
            self, f"{ENDPOINT[device_type]['SETTINGS']}/{setting_type}"
        )

    def process_response(self, json_obj: Dict[str, Any]) -> Dict[str, List[str]]:
        """Return response to command to get list of all setting names and corresponding options for settings of type XList."""
        items = [
            Item(item)
            for item in dict_get_case_insensitive(json_obj, ResponseKey.ITEMS, [])
            if item.type.lower() == TYPE_X_LIST
        ]

        return {item.c_name: item.choices for item in items}


class GetSettingOptionsXListCommand(GetAllSettingsOptionsXListCommand):
    """Command to get options of an audio setting by name (used for setting of type XList)."""

    def __init__(self, device_type: str, setting_type: str, setting_name: str) -> None:
        """Initialize command to get options of an audio setting by name (used for setting of type XList)."""
        self.setting_name = setting_name
        self.setting_type = setting_type.lower()
        super(GetSettingOptionsXListCommand, self).__init__(
            device_type, self.setting_type
        )

    def process_response(self, json_obj: Dict[str, Any]) -> List[str]:
        """Return response to command to get options of an audio setting by name (used for setting of type XList)."""
        return (
            super(GetSettingOptionsXListCommand, self)
            .process_response(json_obj)
            .get(self.setting_name)
        )


class ChangeSettingCommand(ItemCommandBase):
    """Command to set value of a setting by name to new value."""

    def __init__(
        self,
        device_type: str,
        id: int,
        setting_type: str,
        setting_name: str,
        value: Union[int, str],
    ) -> None:
        """Initialize command to set value of a setting by name to new value."""
        super(ChangeSettingCommand, self).__init__(device_type, "SETTINGS", id, value)
        ItemCommandBase.url.fset(
            self, f"{ENDPOINT[device_type]['SETTINGS']}/{setting_type}/{setting_name}"
        )



import logging
import sys
from typing import Union

import click
from tabulate import tabulate

from pyvizio import VizioAsync, guess_device_type
from pyvizio.api.apps import find_app_name
from pyvizio.const import (
    APP_HOME,
    APPS,
    DEFAULT_DEVICE_CLASS,
    DEFAULT_DEVICE_ID,
    DEFAULT_DEVICE_NAME,
    DEFAULT_TIMEOUT,
    DEVICE_CLASS_SPEAKER,
    DEVICE_CLASS_TV,
    NO_APP_RUNNING,
    UNKNOWN_APP,
)
from pyvizio.helpers import async_to_sync
from pyvizio.util import gen_apps_list_from_url

if sys.version_info < (3, 7):
    print("To use this script you need python 3.7 or newer, got %s" % sys.version_info)
    sys.exit(1)

_LOGGER = logging.getLogger(__name__)

pass_vizio = click.make_pass_decorator(VizioAsync)


@click.group(invoke_without_command=False)
@click.option(
    "--ip",
    envvar="VIZIO_IP",
    required=True,
    help=(
        "IP of the device to connect to (optionally add custom port by specifying "
        "'<IP>:<PORT>')"
    ),
    show_default=True,
    show_envvar=True,
)
@click.option(
    "--auth",
    envvar="VIZIO_AUTH",
    required=False,
    default="",
    help=(
        "Auth token for the device to connect to (refer to documentation on how to "
        "obtain auth token)"
    ),
    show_default=True,
    show_envvar=True,
)
@click.option(
    "--device_type",
    envvar="VIZIO_DEVICE_TYPE",
    required=False,
    default=DEFAULT_DEVICE_CLASS,
    type=click.Choice([DEVICE_CLASS_TV, DEVICE_CLASS_SPEAKER]),
    show_default=True,
    show_envvar=True,
)
@click.pass_context
def cli(ctx, ip: str, auth: str, device_type: str) -> None:
    logging.basicConfig(level=logging.INFO)

    ctx.obj = VizioAsync(DEFAULT_DEVICE_ID, ip, DEFAULT_DEVICE_NAME, auth, device_type)


@cli.command()
@click.option(
    "--include_device_type",
    required=False,
    default=False,
    type=click.BOOL,
    help="Include guessed device type (not guaranteed to be correct)",
    show_default=True,
    show_envvar=True,
)
@click.option(
    "--timeout",
    required=False,
    default=DEFAULT_TIMEOUT,
    type=click.IntRange(min=1),
    help="Number of seconds to wait for devices to be discovered",
    show_default=True,
    show_envvar=True,
)
def discover(include_device_type: bool, timeout: int) -> None:
    devices = VizioAsync.discovery_zeroconf(timeout)

    if include_device_type:
        table = tabulate(
            [
                [
                    dev.ip,
                    dev.port,
                    dev.model,
                    dev.name,
                    guess_device_type(dev.ip, dev.port),
                ]
                for dev in devices
            ],
            headers=["IP", "Port", "Model", "Name", "Guessed Device Type"],
        )
    else:
        table = tabulate(
            [[dev.ip, dev.port, dev.model, dev.name] for dev in devices],
            headers=["IP", "Port", "Model", "Name"],
        )

    _LOGGER.info("\n%s", table)


@cli.command()
@async_to_sync
@pass_vizio
async def pair(vizio: VizioAsync) -> None:
    _LOGGER.info("Initiating pairing process, check your device for pin upon success")

    pair_data = await vizio.start_pair()

    if pair_data is not None:
        _LOGGER.info("Challenge type: %s", pair_data.ch_type)
        _LOGGER.info("Challenge token: %s", pair_data.token)
    else:
        _LOGGER.info("ERROR")


@cli.command()
@async_to_sync
@pass_vizio
async def pair_stop(vizio: VizioAsync) -> None:
    _LOGGER.info("Sending stop pair command")

    result = await vizio.stop_pair()

    _LOGGER.info("OK" if result else "ERROR")


@cli.command()
@click.option(
    "--ch_type",
    required=True,
    type=click.INT,
    help="Challenge type obtained from pair command",
    show_default=True,
    show_envvar=True,
)
@click.option(
    "--token",
    required=True,
    type=click.INT,
    help="Challenge token obtained from pair command",
    show_default=True,
    show_envvar=True,
)
@click.option(
    "--pin",
    required=False,
    type=click.STRING,
    default="",
    help="PIN obtained from device after running pair command. Not needed for speaker devices.",
    show_default=True,
    show_envvar=True,
)
@async_to_sync
@pass_vizio
async def pair_finish(vizio: VizioAsync, ch_type: int, token: int, pin: str) -> None:
    _LOGGER.info("Finishing pairing")

    pair_data = await vizio.pair(ch_type, token, pin)

    if pair_data is not None:
        _LOGGER.info("Authorization token: %s", pair_data.auth_token)
    else:
        _LOGGER.info("ERROR")


@cli.command()
@async_to_sync
@pass_vizio
async def get_inputs_list(vizio: VizioAsync) -> None:
    inputs = await vizio.get_inputs_list()

    if inputs:
        table = tabulate(
            [[input.name, input.meta_name] for input in inputs],
            headers=["Name", "Nickname"],
        )

        _LOGGER.info("\n%s", table)
    else:
        _LOGGER.error("Couldn't get list of inputs")


@cli.command()
@async_to_sync
@pass_vizio
async def get_current_input(vizio: VizioAsync) -> None:
    data = await vizio.get_current_input()

    if data is not None:
        _LOGGER.info("Current input: %s", data)
    else:
        _LOGGER.error("Couldn't get current input")


@cli.command()
@async_to_sync
@pass_vizio
async def get_power_state(vizio: VizioAsync) -> None:
    if await vizio.get_power_state():
        _LOGGER.info("Device is on")
    else:
        _LOGGER.info("Device is off")


@cli.command()
@click.argument(
    "state",
    required=False,
    default="toggle",
    type=click.Choice(["toggle", "on", "off"]),
)
@async_to_sync
@pass_vizio
async def power(vizio: VizioAsync, state: str) -> None:
    if "on" == state:
        _LOGGER.info("Turning ON")
        result = await vizio.pow_on()
    elif "off" == state:
        _LOGGER.info("Turning OFF")
        result = await vizio.pow_off()
    else:
        _LOGGER.info("Toggling Power")
        result = await vizio.pow_toggle()

    _LOGGER.info("OK" if result else "ERROR")


@cli.command()
@click.argument(
    "state", required=False, default="up", type=click.Choice(["up", "down"])
)
@click.argument(
    "amount", required=False, default=1, type=click.IntRange(1, 100, clamp=True)
)
@async_to_sync
@pass_vizio
async def volume(vizio: VizioAsync, state: str, amount: int) -> None:
    if "up" == state:
        _LOGGER.info("Increasing volume")
        result = await vizio.vol_up(amount)
    else:
        _LOGGER.info("Decreasing volume")
        result = await vizio.vol_down(amount)

    _LOGGER.info("OK" if result else "ERROR")


@cli.command()
@async_to_sync
@pass_vizio
async def get_volume_level(vizio: VizioAsync) -> None:
    _LOGGER.info("Current volume: %s", await vizio.get_current_volume())


@cli.command()
@pass_vizio
def get_volume_max(vizio: VizioAsync) -> None:
    _LOGGER.info("Max volume: %s", vizio.get_max_volume())


@cli.command()
@click.argument(
    "state",
    required=False,
    default="previous",
    type=click.Choice(["up", "down", "previous"]),
)
@click.argument(
    "amount", required=False, default=1, type=click.IntRange(1, 100, clamp=True)
)
@async_to_sync
@pass_vizio
async def channel(vizio: VizioAsync, state: str, amount: int) -> None:
    if "up" == state:
        _LOGGER.info("Channel up")
        result = await vizio.ch_up(amount)
    elif "down" == state:
        _LOGGER.info("Channel down")
        result = await vizio.ch_down(amount)
    else:
        _LOGGER.info("Previous channel")
        result = await vizio.ch_prev()

    _LOGGER.info("OK" if result else "ERROR")


@cli.command()
@click.argument(
    "state",
    required=False,
    default="toggle",
    type=click.Choice(["toggle", "on", "off"]),
)
@async_to_sync
@pass_vizio
async def mute(vizio: VizioAsync, state: str) -> None:
    if "on" == state:
        _LOGGER.info("Muting")
        result = await vizio.mute_on()
    elif "off" == state:
        _LOGGER.info("Unmuting")
        result = await vizio.mute_off()
    else:
        _LOGGER.info("Toggling mute")
        result = await vizio.mute_toggle()

    _LOGGER.info("OK" if result else "ERROR")


@cli.command()
@async_to_sync
@pass_vizio
async def next_input(vizio: VizioAsync) -> None:
    _LOGGER.info("Circling input")

    result = await vizio.next_input()

    _LOGGER.info("OK" if result else "ERROR")


@cli.command(name="input")
@click.argument("input_name", required=True, type=click.STRING)
@async_to_sync
@pass_vizio
async def input(vizio: VizioAsync, input_name: str) -> None:
    _LOGGER.info("Switching input")

    result = await vizio.set_input(input_name)

    _LOGGER.info("OK" if result else "ERROR")


@cli.command()
@async_to_sync
@pass_vizio
async def play(vizio: VizioAsync) -> None:
    result = await vizio.play()

    _LOGGER.info("OK" if result else "ERROR")


@cli.command()
@async_to_sync
@pass_vizio
async def pause(vizio: VizioAsync) -> None:
    result = await vizio.pause()

    _LOGGER.info("OK" if result else "ERROR")


@cli.command()
@click.argument("key", required=True, type=click.STRING)
@async_to_sync
@pass_vizio
async def key_press(vizio: VizioAsync, key: str) -> None:
    _LOGGER.info("Emulating pressing of '%s' key", key)

    result = await vizio.remote(key.upper())

    _LOGGER.info("OK" if result else "ERROR")


@cli.command()
@pass_vizio
async def get_remote_keys_list(vizio: VizioAsync) -> None:
    table = tabulate(vizio.get_remote_keys_list(), headers=["App Name"])
    _LOGGER.info("\n%s", table)


@cli.command()
@async_to_sync
@pass_vizio
async def get_all_audio_settings(vizio: VizioAsync) -> None:
    audio_settings = await vizio.get_all_audio_settings()
    if audio_settings:
        table = tabulate(
            [[k, v] for k, v in audio_settings.items()], headers=["Name", "Value"]
        )
        _LOGGER.info("\n%s", table)
    else:
        _LOGGER.error("Couldn't get list of audio settings")


@cli.command()
@async_to_sync
@pass_vizio
async def get_all_audio_settings_options(vizio: VizioAsync) -> None:
    audio_settings_options = await vizio.get_all_audio_settings_options()
    if audio_settings_options:
        options = []
        for k, v in audio_settings_options.items():
            if isinstance(v, dict):
                options.append([k, v.get("default"), v["min"], v["max"], None])
            else:
                options.append([k, None, None, None, ", ".join(v)])
        table = tabulate(options, headers=["Name", "Default", "Min", "Max", "Choices"])
        _LOGGER.info("\n%s", table)
    else:
        _LOGGER.error("Couldn't get list of audio settings options")


@cli.command()
@click.argument("setting_name", required=True, type=click.STRING)
@async_to_sync
@pass_vizio
async def get_audio_setting(vizio: VizioAsync, setting_name: str) -> None:
    value = await vizio.get_audio_setting(setting_name)

    if value is not None:
        _LOGGER.info("Current '%s' setting: %s", setting_name, value)
    else:
        _LOGGER.error("Couldn't get value for '%s' setting", setting_name)


@cli.command()
@click.argument("setting_type", required=True, type=click.STRING)
@click.argument("setting_name", required=True, type=click.STRING)
@async_to_sync
@pass_vizio
async def get_setting_options(
    vizio: VizioAsync, setting_type: str, setting_name: str
) -> None:
    value = await vizio.get_setting_options(setting_type, setting_name)

    _LOGGER.error(value)
    if value is not None:
        if isinstance(value, dict):
            if value.get("default") is not None:
                table = tabulate(
                    [[value["default"], value["min"], value["max"]]],
                    headers=["Default", "Min", "Max"],
                )
            else:
                table = tabulate([[value["min"], value["max"]]], headers=["Min", "Max"])
            _LOGGER.info("For '%s' setting:\n%s", setting_name, table)
        else:
            _LOGGER.info("Options for '%s' setting: %s", setting_name, ", ".join(value))
    else:
        _LOGGER.error("Couldn't get options for '%s' setting", setting_name)


@cli.command()
@click.argument("setting_type", required=True, type=click.STRING)
@click.argument("setting_name", required=True, type=click.STRING)
@click.argument("new_value", required=True, type=click.STRING)
@async_to_sync
@pass_vizio
async def setting(
    vizio: VizioAsync, setting_type: str, setting_name: str, new_value: Union[int, str]
) -> None:
    _LOGGER.info("Attempting to set '%s' to '%s'", setting_name, new_value)

    try:
        result = await vizio.set_setting(setting_type, setting_name, int(new_value))
    except ValueError:
        result = await vizio.set_setting(setting_type, setting_name, new_value)

    _LOGGER.info("OK" if result else "ERROR")


@cli.command()
@click.argument("setting_type", required=True, type=click.STRING)
@async_to_sync
@pass_vizio
async def get_all_settings(vizio: VizioAsync, setting_type: str) -> None:
    settings = await vizio.get_all_settings(setting_type)
    if settings:
        table = tabulate(
            [[k, v] for k, v in settings.items()], headers=["Name", "Value"]
        )
        _LOGGER.info("\n%s", table)
    else:
        _LOGGER.error("Couldn't get list of settings for %s setting type", setting_type)


@cli.command()
@click.argument("setting_type", required=True, type=click.STRING)
@async_to_sync
@pass_vizio
async def get_all_settings_options(vizio: VizioAsync, setting_type: str) -> None:
    settings_options = await vizio.get_all_settings_options(setting_type)
    if settings_options:
        options = []
        for k, v in settings_options.items():
            if isinstance(v, dict):
                options.append([k, v.get("default"), v["min"], v["max"], None])
            else:
                options.append([k, None, None, None, ", ".join(v)])
        table = tabulate(options, headers=["Name", "Default", "Min", "Max", "Choices"])
        _LOGGER.info("\n%s", table)
    else:
        _LOGGER.error(
            "Couldn't get list of settings options for %s setting type", setting_type
        )


@cli.command()
@click.argument("setting_type", required=True, type=click.STRING)
@click.argument("setting_name", required=True, type=click.STRING)
@async_to_sync
@pass_vizio
async def get_setting(vizio: VizioAsync, setting_type: str, setting_name: str) -> None:
    value = await vizio.get_setting(setting_type, setting_name)

    if value is not None:
        _LOGGER.info("Current '%s' setting: %s", setting_name, value)
    else:
        _LOGGER.error(
            "Couldn't get value for '%s' setting of %s setting type",
            setting_name,
            setting_type,
        )


@cli.command()
@async_to_sync
@pass_vizio
async def get_setting_types_list(vizio: VizioAsync) -> None:
    value = await vizio.get_setting_types_list()

    if value is not None:
        table = tabulate(
            [[setting_type] for setting_type in value], headers=["Setting Type"]
        )
        _LOGGER.info("\n%s", table)
    else:
        _LOGGER.error("Couldn't get setting types")


@cli.command()
@click.argument("setting_name", required=True, type=click.STRING)
@async_to_sync
@pass_vizio
async def get_audio_setting_options(vizio: VizioAsync, setting_name: str) -> None:
    value = await vizio.get_audio_setting_options(setting_name)

    _LOGGER.error(value)
    if value is not None:
        if isinstance(value, dict):
            if value.get("default") is not None:
                table = tabulate(
                    [[value["default"], value["min"], value["max"]]],
                    headers=["Default", "Min", "Max"],
                )
            else:
                table = tabulate([[value["min"], value["max"]]], headers=["Min", "Max"])
            _LOGGER.info("For '%s' setting:\n%s", setting_name, table)
        else:
            _LOGGER.info("Options for '%s' setting: %s", setting_name, ", ".join(value))
    else:
        _LOGGER.error("Couldn't get options for '%s' setting", setting_name)


@cli.command()
@click.argument("setting_name", required=True, type=click.STRING)
@click.argument("new_value", required=True, type=click.STRING)
@async_to_sync
@pass_vizio
async def audio_setting(
    vizio: VizioAsync, setting_name: str, new_value: Union[int, str]
) -> None:
    _LOGGER.info("Attempting to set '%s' to '%s'", setting_name, new_value)

    try:
        result = await vizio.set_audio_setting(setting_name, int(new_value))
    except ValueError:
        result = await vizio.set_audio_setting(setting_name, new_value)

    _LOGGER.info("OK" if result else "ERROR")


@cli.command()
@click.option(
    "--country",
    required=False,
    type=click.Choice(["usa", "can", "mexico", "all"]),
    default="all",
    help="Only apps supported in the specified country will be returned",
    show_default=True,
    show_envvar=True,
)
@async_to_sync
@pass_vizio
async def get_apps_list(vizio: VizioAsync, country: str = "all") -> None:
    apps = await vizio.get_apps_list(country)
    if apps:
        table = tabulate([{"Name": app} for app in apps], headers="keys")
        _LOGGER.info("\n%s", table)
    else:
        _LOGGER.error("Couldn't get list of apps")


@cli.command()
@click.argument("app_name", required=True, type=click.STRING)
@async_to_sync
@pass_vizio
async def launch_app(vizio: VizioAsync, app_name: str) -> None:
    _LOGGER.info("Attempting to launch '%s' app", app_name)
    apps_list = await gen_apps_list_from_url()
    if not apps_list:
        apps_list = APPS

    result = await vizio.launch_app(app_name, apps_list)

    _LOGGER.info("OK" if result else "ERROR")


@cli.command()
@click.argument("APP_ID", required=True, type=click.STRING)
@click.argument("NAME_SPACE", required=True, type=click.IntRange(min=0))
@click.argument("MESSAGE", required=False, type=click.STRING, default=None)
@async_to_sync
@pass_vizio
async def launch_app_config(
    vizio: VizioAsync, APP_ID: str, NAME_SPACE: int, MESSAGE: str
) -> None:
    _LOGGER.info(
        "Attempting to launch app using config %s",
        {"APP_ID": APP_ID, "NAME_SPACE": NAME_SPACE, "MESSAGE": MESSAGE},
    )

    result = await vizio.launch_app_config(APP_ID, NAME_SPACE, MESSAGE)

    _LOGGER.info("OK" if result else "ERROR")


@cli.command()
@async_to_sync
@pass_vizio
async def get_current_app(vizio: VizioAsync) -> None:
    app_config = await vizio.get_current_app_config()
    apps_list = await gen_apps_list_from_url()
    if not apps_list:
        apps_list = APPS
    app_name = find_app_name(app_config, [APP_HOME, *apps_list])

    if app_name:
        if app_name == NO_APP_RUNNING:
            _LOGGER.info("No currently running app")
        elif app_name == UNKNOWN_APP:
            _LOGGER.info(
                "Can't determine the name of the app, the currently running app's config is %s",
                app_config,
            )
        else:
            _LOGGER.info("Currently running app: %s", app_name)
    else:
        _LOGGER.error("Couldn't get currently running app")


@cli.command()
@async_to_sync
@pass_vizio
async def get_current_app_config(vizio: VizioAsync) -> None:
    app_config = await vizio.get_current_app_config()

    if app_config is None:
        _LOGGER.error("Couldn't get currently running app")
    elif app_config:
        _LOGGER.info("Currently running app's config: %s", app_config)
    else:
        _LOGGER.info("No currently running app")


if __name__ == "__main__":
    cli()



"""Vizio SmartCast API commands and class for device inputs."""

from typing import Any, Dict, List, Optional

from pyvizio.api._protocol import ResponseKey
from pyvizio.api.item import Item, ItemCommandBase, ItemInfoCommandBase
from pyvizio.helpers import dict_get_case_insensitive


class InputItem(Item):
    """Input device."""

    def __init__(self, json_item: Dict[str, Any], is_extended_metadata: bool) -> None:
        """Initialize input device."""
        super(InputItem, self).__init__(json_item)
        self.meta_name = None
        self.meta_data = None

        meta = dict_get_case_insensitive(json_item, ResponseKey.VALUE)

        if meta:
            if is_extended_metadata:
                self.meta_name = dict_get_case_insensitive(meta, ResponseKey.NAME)
                self.meta_data = dict_get_case_insensitive(meta, ResponseKey.METADATA)
            else:
                self.meta_name = meta

        if not self.meta_name:
            self.meta_name = self.c_name


class GetInputsListCommand(ItemInfoCommandBase):
    """Command to get list of available inputs."""

    def __init__(self, device_type: str) -> None:
        """Initialize command to get list of available inputs."""
        super(GetInputsListCommand, self).__init__(device_type, "INPUTS")

    def process_response(self, json_obj: Dict[str, Any]) -> Optional[List[InputItem]]:
        """Return response to command to get list of available inputs."""
        items = dict_get_case_insensitive(json_obj, ResponseKey.ITEMS)

        if items:
            return [
                InputItem(itm, True)
                for itm in items
                if dict_get_case_insensitive(itm, ResponseKey.CNAME) != "current_input"
            ]

        return None


class GetCurrentInputCommand(ItemInfoCommandBase):
    """Command to get currently active input."""

    def __init__(self, device_type: str) -> None:
        """Initialize command to get currently active input."""
        super(GetCurrentInputCommand, self).__init__(device_type, "CURRENT_INPUT")

    def process_response(self, json_obj: Dict[str, Any]) -> Optional[InputItem]:
        """Return response to command to get currently active input."""
        items = dict_get_case_insensitive(json_obj, ResponseKey.ITEMS)

        v_input = None

        if items:
            v_input = InputItem(items[0], False)

        return v_input


class ChangeInputCommand(ItemCommandBase):
    """Command to change active input by name."""

    def __init__(self, device_type: str, id: int, name: str) -> None:
        """Initialize command to change active input by name."""
        super(ChangeInputCommand, self).__init__(device_type, "CURRENT_INPUT", id, name)



"""Vizio SmartCast API commands for apps."""

from typing import Any, Dict, List, Union

from pyvizio.api._protocol import ENDPOINT, ResponseKey
from pyvizio.api.base import CommandBase
from pyvizio.api.input import ItemInfoCommandBase
from pyvizio.const import (
    APP_CAST,
    APP_HOME,
    EQUIVALENT_NAME_SPACES,
    NO_APP_RUNNING,
    UNKNOWN_APP,
)
from pyvizio.helpers import dict_get_case_insensitive


class AppConfig(object):
    """Vizio SmartCast app config."""

    def __init__(
        self, APP_ID: str = None, NAME_SPACE: int = None, MESSAGE: str = None
    ) -> None:
        self.APP_ID = APP_ID
        self.NAME_SPACE = NAME_SPACE
        self.MESSAGE = MESSAGE

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.__dict__})"

    def __eq__(self, other) -> bool:
        return self is other or self.__dict__ == other.__dict__

    def __bool__(self) -> bool:
        return self != AppConfig()


def find_app_name(config_to_check: AppConfig, app_list: List[Dict[str, Any]]) -> str:
    """
    Return the app name for a given AppConfig based on a list of apps.

    Returns UNKNOWN_APP if app name can't be found in APPS list for given AppConfig.
    """
    if not config_to_check:
        return NO_APP_RUNNING

    # Attempt to find an exact match from known apps list
    for app_def in app_list:
        if isinstance(app_def["config"], list):
            for config in app_def["config"]:
                if (
                    config["APP_ID"] == config_to_check.APP_ID
                    and config["NAME_SPACE"] == config_to_check.NAME_SPACE
                ):
                    return app_def["name"]
        elif (
            isinstance(app_def["config"], dict)
            and app_def["config"]["APP_ID"] == config_to_check.APP_ID
            and app_def["config"]["NAME_SPACE"] == config_to_check.NAME_SPACE
        ):
            return app_def["name"]

    # If exact match couldn't be find, swap in equivalent name spaces
    # and attempt to find a match
    if config_to_check.NAME_SPACE in EQUIVALENT_NAME_SPACES:
        for app_def in app_list:
            if isinstance(app_def["config"], list):
                for config in app_def["config"]:
                    if (
                        config["APP_ID"] == config_to_check.APP_ID
                        and config["NAME_SPACE"] in EQUIVALENT_NAME_SPACES
                    ):
                        return app_def["name"]
            elif (
                isinstance(app_def["config"], dict)
                and app_def["config"]["APP_ID"] == config_to_check.APP_ID
                and app_def["config"]["NAME_SPACE"] in EQUIVALENT_NAME_SPACES
            ):
                return app_def["name"]

    # So far only the SmartCast home screen appears to use the NAME_SPACE of 0
    if config_to_check.NAME_SPACE == 0:
        return APP_CAST

    # If no match, app is unknown
    return UNKNOWN_APP


class LaunchAppConfigCommand(CommandBase):
    """Command to launch app by config."""

    def __init__(
        self, device_type: str, APP_ID: str, NAME_SPACE: int, MESSAGE: str = None
    ) -> None:
        """Initialize command to launch app by config."""
        super(LaunchAppConfigCommand, self).__init__(
            ENDPOINT[device_type]["LAUNCH_APP"]
        )

        self.VALUE = AppConfig(APP_ID, NAME_SPACE, MESSAGE)


class LaunchAppNameCommand(LaunchAppConfigCommand):
    """Command to launch app by name."""

    def __init__(
        self,
        device_type: str,
        app_name: str,
        apps_list: List[Dict[str, Union[str, List[Union[str, Dict[str, Any]]]]]],
    ) -> None:
        """Initialize command to launch app by name."""
        app_def = next(
            (
                app_def
                for app_def in [APP_HOME, *apps_list]
                if app_def["name"].lower() == app_name.lower()
            ),
            dict(),
        )

        # Unpack config dict into expected key/value argument pairs
        super(LaunchAppNameCommand, self).__init__(
            device_type, **app_def.get("config", [{}])[0]
        )


class GetCurrentAppConfigCommand(ItemInfoCommandBase):
    """Command to get currently running app's config."""

    def __init__(self, device_type: str) -> None:
        """Initialize command to get currently running app's config."""
        super(GetCurrentAppConfigCommand, self).__init__(device_type, "CURRENT_APP")

    def process_response(self, json_obj: Dict[str, Any]) -> AppConfig:
        """Return response to command to get currently running app's config."""
        item = dict_get_case_insensitive(json_obj, ResponseKey.ITEM, {})
        current_app_id = dict_get_case_insensitive(item, ResponseKey.VALUE)

        if current_app_id:
            return AppConfig(**current_app_id)

        return AppConfig()


class GetCurrentAppNameCommand(GetCurrentAppConfigCommand):
    """Command to get currently running app's name."""

    def __init__(
        self,
        device_type: str,
        apps_list: List[Dict[str, Union[str, List[Union[str, Dict[str, Any]]]]]],
    ) -> None:
        """Initialize command to get currently running app's name."""
        super(GetCurrentAppNameCommand, self).__init__(device_type)
        self.apps_list = apps_list

    def process_response(self, json_obj: Dict[str, Any]) -> str:
        """
        Return response to command to get currently running app's name.

        Returns NO_APP_RUNNING if no app is currently running and UNKNOWN_APP
        if app name can't be retrieved from APPS.
        """
        current_app_config = super(GetCurrentAppNameCommand, self).process_response(
            json_obj
        )

        if current_app_config:
            return find_app_name(current_app_config, [APP_HOME, *self.apps_list])

        # Return NO_APP_RUNNING if value from response was None
        return NO_APP_RUNNING
